from collections import defaultdict
from typing import Dict, Set, NamedTuple, Tuple


class DataItem(NamedTuple):
    """
    A DataItem refers to an item in an object store, for example an S3 (simple storage service) object.
    """
    bucket: str
    name: str
    size: int


class StorageIndex:
    """
    The StorageIndex keeps track of data items on a set of storage nodes in the cluster.
    Currently this is only a dummy in-memory implementation.
    """
    buckets: Dict[str, Set[str]]
    tree: Dict[Tuple[str, str], Set[str]]
    items: Dict[Tuple[str, str], DataItem]

    def __init__(self) -> None:
        super().__init__()
        self.buckets = defaultdict(set)
        self.tree = defaultdict(set)
        self.items = dict()

    def mb(self, name: str, node: str):
        """
        Create a bucket on a node.

        :param name: the bucket name
        :param node: the node to create the bucket on
        """
        self.buckets[name].add(node)

    def put(self, data: DataItem):
        nodes = self.get_bucket_nodes(data.bucket)
        if not nodes:
            raise KeyError('no nodes that host bucket %s' % data.bucket)

        k = (data.bucket, data.name)
        self.items[k] = data

        for node in nodes:
            self.tree[k].add(node)

    def stat(self, bucket: str, name: str) -> DataItem:
        k = (bucket, name)
        return self.items.get(k)

    def get_bucket_nodes(self, bucket: str) -> Set[str]:
        return self.buckets[bucket]

    def get_data_nodes(self, bucket: str, name: str) -> Set[str]:
        k = (bucket, name)
        return self.tree.get(k)

    def print_ls_tree(self):
        """
        Debug function to print the entire tree in a `find`-style output.
        """
        tree = defaultdict(lambda: defaultdict(list))
        for (bucket, item), nodes in self.tree.items():
            for node in nodes:
                tree[node][bucket].append(item)

        for node, buckets in tree.items():
            print(f'/{node}')

            for bucket, items in buckets.items():
                print(f'/{node}/{bucket}')

                for item in items:
                    print(f'/{node}/{bucket}/{item}')



"""
Implementations of the kubernetes default scheduler's predicates.
https://github.com/kubernetes/kubernetes/blob/e318642946daab9e0330757a3556a1913bb3fc5c/pkg/scheduler/algorithm/predicates/
"""
import logging
from typing import List

from skippy.core.clustercontext import ClusterContext
from skippy.core.model import Pod, Node, Capacity

logger = logging.getLogger(__name__)


class Predicate:
    """
    Abstract class for predicate implementations.
    """

    def __init__(self):
        pass

    def passes_predicate(self, context: ClusterContext, pod: Pod, node: Node) -> bool:
        raise NotImplementedError


class CombinedPredicate(Predicate):
    """
    Helper-Super-Class to combine multiple predicates to a conjunction.
    """

    def __init__(self, predicates: [Predicate]):
        super().__init__()
        self.predicates = predicates

    def passes_predicate(self, context: ClusterContext, pod: Pod, node: Node) -> bool:
        return all(self.__passes_and_logs_predicate(predicate, context, pod, node)
                   for predicate in self.predicates)

    # noinspection PyMethodMayBeStatic
    def __passes_and_logs_predicate(self, predicate: Predicate, context: ClusterContext, pod: Pod, node: Node):
        result = predicate.passes_predicate(context, pod, node)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'Pod {pod.name} / Node {node.name} / {type(predicate).__name__}: '
                         f'{"Passed" if result else "Failed"}')

        return result


class PodFitsResourcesPred(Predicate):
    """
    Part of NonCriticalPreds!
    PodFitsResources checks if a node has sufficient resources, such as cpu, memory, gpu, opaque int resources etc to
    run a pod.
    https://github.com/kubernetes/kubernetes/blob/eaa78b88ac25a61bfb1aa81d118c5ffeda041b64/pkg/scheduler/algorithm/predicates/predicates.go#L769
    """

    def passes_predicate(self, context: ClusterContext, pod: Pod, node: Node) -> bool:
        allocatable = node.allocatable
        requested = Capacity(0, 0)
        for container in pod.spec.containers:
            requested.cpu_millis += container.resources.requests.get('cpu', container.resources.
                                                                     default_milli_cpu_request)
            requested.memory += container.resources.requests.get('memory', container.resources.default_mem_request)
        passed = requested.memory <= allocatable.memory and requested.cpu_millis <= allocatable.cpu_millis

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'Pod {pod.name} requests {requested.cpu_millis} / {requested.memory}. '
                         f'Available on node {node.name}: {allocatable.cpu_millis} / {allocatable.memory}.'
                         f'Passed: {passed}')
        return passed


class NonCriticalPreds(CombinedPredicate):
    """
    Part of GeneralPreds!
    NoncriticalPredicates are the predicates that only non-critical pods need
    https://github.com/kubernetes/kubernetes/blob/eaa78b88ac25a61bfb1aa81d118c5ffeda041b64/pkg/scheduler/algorithm/predicates/predicates.go#L1134
    """

    def __init__(self):
        super().__init__([PodFitsResourcesPred()])


class EssentialPreds(CombinedPredicate):
    """
    Part of GeneralPreds!
    EssentialPredicates are the predicates that all pods, including critical pods, need
    https://github.com/kubernetes/kubernetes/blob/eaa78b88ac25a61bfb1aa81d118c5ffeda041b64/pkg/scheduler/algorithm/predicates/predicates.go#L1148
    """

    def __init__(self):
        super().__init__([PodFitsResourcesPred()])


class GeneralPreds(CombinedPredicate):
    """
    GeneralPredicates checks whether noncriticalPredicates and EssentialPredicates pass.
    noncriticalPredicates are the predicates that only non-critical pods need and EssentialPredicates are the predicates
    that all pods, including critical pods, need
    https://github.com/kubernetes/kubernetes/blob/eaa78b88ac25a61bfb1aa81d118c5ffeda041b64/pkg/scheduler/algorithm/predicates/predicates.go#L1110
    """

    def __init__(self):
        # NonCriticalPreds should only be applied if the Pod is non-critical,
        # but we don't handle critical pods in our simulation
        super().__init__([EssentialPreds(), NonCriticalPreds()])


class CheckNodeLabelPresencePred(Predicate):
    def __init__(self, labels: List[str], should_be_present=True) -> None:
        super().__init__()
        self.labels = labels
        self.should_be_present = should_be_present

        if should_be_present:
            self._passes_predicate = self.has_labels
        else:
            self._passes_predicate = self.has_labels_not

    def passes_predicate(self, context: ClusterContext, pod: Pod, node: Node) -> bool:
        return self._passes_predicate(node)

    def has_labels(self, node: Node) -> bool:
        for label in self.labels:
            if label not in node.labels:
                return False

        return True

    def has_labels_not(self, node: Node) -> bool:
        for label in self.labels:
            if label in node.labels:
                return False

        return True



import unittest

from skippy.core.utils import parse_size_string


class ParseSizeStringTest(unittest.TestCase):
    def test_parse_size_string(self):
        self.assertEqual(1, parse_size_string('1'))
        self.assertEqual(1_000, parse_size_string('1K'))
        self.assertEqual(1_000_000, parse_size_string('1M'))
        self.assertEqual(1_048_576, parse_size_string('1Mi'))

    def test_parse_size_string_error(self):
        self.assertRaises(ValueError, parse_size_string, 'foo')
        self.assertRaises(ValueError, parse_size_string, '')



import logging
from itertools import islice, cycle
from operator import itemgetter, add
from typing import List, Tuple

from skippy.core.clustercontext import ClusterContext
from skippy.core.model import Pod, Node, SchedulingResult
from skippy.core.predicates import Predicate, PodFitsResourcesPred, CheckNodeLabelPresencePred
from skippy.core.priorities import Priority, BalancedResourcePriority, \
    LatencyAwareImageLocalityPriority, CapabilityPriority, DataLocalityPriority, LocalityTypePriority
from skippy.core.utils import normalize_image_name

logger = logging.getLogger(__name__)


class Scheduler:
    # Needs to contain all predicates that should be executed (if they're not overwritten in the constructor)
    default_predicates: List[Predicate] = [
        PodFitsResourcesPred(),
        CheckNodeLabelPresencePred(['data.skippy.io/storage'], False)  # do not schedule functions on storage nodes
    ]

    # Needs to contain all priorities that should be executed (if they're not overwritten in the constructor)
    default_priorities: List[Tuple[float, Priority]] = [(1.0, BalancedResourcePriority()),
                                                        (1.0, LatencyAwareImageLocalityPriority()),
                                                        (1.0, LocalityTypePriority()),
                                                        (1.0, DataLocalityPriority()),
                                                        (1.0, CapabilityPriority())]

    # https://github.com/kubernetes/kubernetes/blob/c1f40a5310b0abfe9a4fbddc24955360821a324b/pkg/scheduler/core/generic_scheduler.go#L58
    min_feasible_nodes_to_find = 100

    # https://github.com/kubernetes/kubernetes/blob/c1f40a5310b0abfe9a4fbddc24955360821a324b/pkg/scheduler/core/generic_scheduler.go#L63
    min_feasible_nodes_percentage_to_find = 5

    # https://github.com/kubernetes/kubernetes/blob/c1f40a5310b0abfe9a4fbddc24955360821a324b/pkg/scheduler/api/types.go#L40
    default_percentage_of_nodes_to_score = 50

    def __init__(self, cluster_context: ClusterContext, percentage_of_nodes_to_score: int = 100,
                 predicates: List[Predicate] = None,
                 priorities: List[Tuple[float, Priority]] = None):
        if priorities is None:
            priorities = self.default_priorities
        if predicates is None:
            predicates = self.default_predicates

        self.predicates = predicates
        self.priorities = priorities
        self.percentage_of_nodes_to_score = percentage_of_nodes_to_score

        # Context containing all the cluster information
        self.cluster_context = cluster_context

        # Defines at which index the last scoring stopped (i.e. where the next one should start)
        self.last_scored_node_index = 0

    def schedule(self, pod: Pod) -> SchedulingResult:
        """
        Decides on which node to place a pod.

        :param pod: to place
        :return: str name of the node to place the pod on
        """
        logging.debug('Received a new pod to schedule: %s', pod.name)

        nodes = self.cluster_context.list_nodes()
        num_of_nodes_to_find = self.__num_feasible_nodes_to_find(len(nodes))

        filtered = filter(lambda node: self.passes_predicates(pod, node),
                          islice(cycle(nodes), self.last_scored_node_index, self.last_scored_node_index + len(nodes)))
        feasible_nodes: [Node] = list(islice(filtered, num_of_nodes_to_find))
        if len(feasible_nodes) > 0:
            self.last_scored_node_index = (nodes.index(feasible_nodes[-1]) + 1) % len(nodes)

        cluster = self.cluster_context

        # Score all feasible nodes
        # Possible: The generic_scheduler.go parallelizes the score calculation (map reduce pattern)
        # We could just use multiprocessing.Pool()'s map function?
        # https://stackoverflow.com/questions/1704401/is-there-a-simple-process-based-parallel-map-for-python
        # TODO this loop could be heavily optimized, especially when removing the priority reduction step
        scored_nodes: [int] = [0] * len(feasible_nodes)
        for weighted_priority in self.priorities:
            weight = weighted_priority[0]
            function = weighted_priority[1]
            mapped_nodes = [function.map_node_score(cluster, pod, node) for node in feasible_nodes]
            reduced_node_scores = function.reduce_mapped_score(cluster, pod, feasible_nodes, mapped_nodes)
            weighted_node_scores = [score * weight for score in reduced_node_scores]
            scored_nodes = list(map(add, weighted_node_scores, scored_nodes))

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Pod %s / %s: %s', pod.name, type(function), weighted_node_scores)

        scored_named_nodes: [(Node, int)] = list(zip(feasible_nodes, scored_nodes))

        logging.debug('Node scores: %s', scored_named_nodes)

        # Find the name of the node with the highest score or None
        sorted_scored_nodes = max(scored_named_nodes, key=itemgetter(1), default=(None, 0))
        suggested_host: Node = next(iter(sorted_scored_nodes), None)
        needed_images = None

        if suggested_host is not None:
            # Add a list of images needed to pull to the result (before manipulating the state with #place_pod_on_node
            needed_images = []
            host_images = self.cluster_context.images_on_nodes[suggested_host.name]
            for container in pod.spec.containers:
                if normalize_image_name(container.image) not in host_images:
                    needed_images.append(normalize_image_name(container.image))

            self.cluster_context.place_pod_on_node(pod, suggested_host)
            logging.debug('Found best node. Remaining allocatable resources after scheduling: %s',
                          suggested_host.allocatable)

        return SchedulingResult(suggested_host=suggested_host, feasible_nodes=len(feasible_nodes),
                                needed_images=needed_images)

    def passes_predicates(self, pod: Pod, node: Node) -> bool:
        # Conjunction over all node predicate checks
        return all(self.__passes_and_logs_predicate(predicate, self.cluster_context, pod, node)
                   for predicate in self.predicates)

    # noinspection PyMethodMayBeStatic
    def __passes_and_logs_predicate(self, predicate: Predicate, context: ClusterContext, pod: Pod, node: Node):
        result = predicate.passes_predicate(context, pod, node)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'Pod {pod.name} / Node {node.name} / {type(predicate).__name__}: '
                         f'{"Passed" if result else "Failed"}')
        return result

    # noinspection PyMethodMayBeStatic
    def __num_feasible_nodes_to_find(self, num_all_nodes: int) -> int:
        """
        Calculates the number of nodes which should be scored by the scheduler (changed in K8s 1.14):
        https://github.com/kubernetes/kubernetes/blob/c1f40a5310b0abfe9a4fbddc24955360821a324b/pkg/scheduler/core/generic_scheduler.go#L441
        https://kubernetes.io/docs/concepts/scheduling/scheduler-perf-tuning/#percentage-of-nodes-to-score
        :param num_all_nodes: total number of nodes
        :return: number of nodes the scheduler should score
        """
        if num_all_nodes < self.min_feasible_nodes_percentage_to_find or self.percentage_of_nodes_to_score >= 100:
            return num_all_nodes
        adaptive_percentage: float = self.percentage_of_nodes_to_score
        if adaptive_percentage <= 0:
            adaptive_percentage = self.default_percentage_of_nodes_to_score - num_all_nodes / 125
            if adaptive_percentage < self.min_feasible_nodes_percentage_to_find:
                adaptive_percentage = self.min_feasible_nodes_percentage_to_find
        num_nodes = int(num_all_nodes * adaptive_percentage / 100)
        if num_nodes < self.min_feasible_nodes_to_find:
            return self.min_feasible_nodes_to_find
        return num_nodes



import re
import time

"""
https://github.com/kubernetes/kubernetes/blob/e318642946daab9e0330757a3556a1913bb3fc5c/pkg/util/parsers/parsers.go#L30
"""
default_image_tag: str = "latest"


def normalize_image_name(image_name: str):
    """
    normalizedImageName returns the CRI compliant name for a given image.

    https://github.com/kubernetes/kubernetes/blob/e318642946daab9e0330757a3556a1913bb3fc5c/pkg/scheduler/algorithm/priorities/image_locality.go#L104

    :param image_name: the full container image name
    :return: the normalized name
    """
    if image_name.rfind(":") <= image_name.rfind("/"):
        image_name = image_name + ":" + default_image_tag
    return image_name


__size_conversions = {
    'K': 10 ** 3,
    'M': 10 ** 6,
    'G': 10 ** 9,
    'T': 10 ** 12,
    'P': 10 ** 15,
    'E': 10 ** 18,
    'Ki': 2 ** 10,
    'Mi': 2 ** 20,
    'Gi': 2 ** 30,
    'Ti': 2 ** 40,
    'Pi': 2 ** 50,
    'Ei': 2 ** 60
}

__size_pattern = re.compile(r"([0-9]+)([a-zA-Z]*)")


def parse_size_string(size_string: str) -> int:
    m = __size_pattern.match(size_string)
    if not m:
        raise ValueError('invalid size string: %s' % size_string)

    if len(m.groups()) > 1:
        number = m.group(1)
        unit = m.group(2)
        return int(number) * __size_conversions.get(unit, 1)
    else:
        return int(m.group(1))


class Timer:

    def __init__(self) -> None:
        super().__init__()
        self.then = -1

    def start(self):
        self.then = time.time()
        return self

    def ms(self):
        return (time.time() - self.then) * 1000


def counter(start: int = 1):
    n = start
    while True:
        yield n
        n += 1



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Dict

from skippy.core.model import Node, Pod, ImageState
from skippy.core.storage import StorageIndex
from skippy.core.utils import normalize_image_name

BandwidthGraph = Dict[str, Dict[str, float]]


class ClusterContext(ABC):
    """
    The cluster context holds runtime information about the cluster. In K8s, this is stored in etcd and made available
    through the API server.
    """

    def __init__(self):
        # Dict holding image metadata for each (normalized) image tag
        self.image_states: Dict[str, ImageState] = self.get_init_image_states()

        # Defines the maximum score of a single node
        self.max_priority: int = 10

        # Dict to maintain which node has which images in the local registry
        self.images_on_nodes: Dict[str, Dict[str, ImageState]] = defaultdict(dict)

        # Dict to maintain the bandwidth graph
        # bandwidth[from][to] = bandwidth in bytes per second
        self.bandwidth: BandwidthGraph = self.get_bandwidth_graph()

        self.storage_index: StorageIndex = None

    def get_node(self, name: str) -> Node:
        for node in self.list_nodes():
            if node.name == name:
                return node

    @abstractmethod
    def get_init_image_states(self) -> Dict[str, ImageState]:
        raise NotImplemented()

    @abstractmethod
    def get_bandwidth_graph(self) -> Dict[str, Dict[str, float]]:
        raise NotImplemented()

    @abstractmethod
    def list_nodes(self) -> List[Node]:
        raise NotImplemented()

    @abstractmethod
    def get_next_storage_node(self, node: Node) -> str:
        raise NotImplemented()

    def get_storage_nodes(self, urn: str) -> List[str]:
        """
        Return a list of storage nodes (MinIO pods) that hold the data item with the given URN.
        Currently the URN is simply am S3 path string <bucket>/<item>
        :param urn: S3 path string <bucket>/<item>
        :return: a list of nodes that hold the data item
        """
        bucket, name = urn.split('/')  # TODO: proper addressing scheme
        # FIXME storage: currently we assume that every bucket lives on a single node
        return [name for name in self.storage_index.get_bucket_nodes(bucket)]

    def place_pod_on_node(self, pod: Pod, node: Node):
        """
        Method to keep track of already placed pods on nodes in order to allow calculating the remaining resources on a
        node.
        """
        for container in pod.spec.containers:
            image_name = normalize_image_name(container.image)

            if image_name not in self.images_on_nodes[node.name]:
                image_state = self.get_image_state(image_name)

                image_state.num_nodes += 1

                images_on_nodes = self.images_on_nodes[node.name]
                images_on_nodes[image_name] = image_state
                self.images_on_nodes[node.name][image_name] = image_state  # FIXME: isn't this the same statement?

            required_cpu_millis = container.resources.requests.get('cpu', container.resources.default_milli_cpu_request)
            required_memory = container.resources.requests.get('memory', container.resources.default_mem_request)

            node.allocatable.cpu_millis -= required_cpu_millis
            node.allocatable.memory -= required_memory
        node.pods.append(pod)

    def remove_pod_from_node(self, pod: Pod, node: Node):
        for container in pod.spec.containers:
            required_cpu_millis = container.resources.requests.get('cpu', container.resources.default_milli_cpu_request)
            required_memory = container.resources.requests.get('memory', container.resources.default_mem_request)

            node.allocatable.cpu_millis += required_cpu_millis
            node.allocatable.memory += required_memory
        node.pods.remove(pod)

    def remove_pod_images_from_node(self, pod: Pod, node: Node):
        for container in pod.spec.containers:
            image_name = normalize_image_name(container.image)

            if image_name in self.images_on_nodes[node.name]:
                image_state = self.get_image_state(image_name)
                image_state.num_nodes -= 1
                del self.images_on_nodes[node.name][image_name]

    def get_image_state(self, image_name: str) -> ImageState:
        """
        Finds metadata about the image.
        :returns ImageState
        """
        if self.image_states.get(image_name) is None:
            self.image_states[image_name] = self.retrieve_image_state(image_name)
        return self.image_states[image_name]

    def retrieve_image_state(self, image_name):
        # TODO maybe implement docker integration? There's no proper documented API, but f.e.
        #  https://cloud.docker.com/v2/repositories/alexrashed/ml-wf-1-pre/tags/0.33/
        #  returns a JSON containing the size
        raise NotImplemented("Remote requested size information about images are not yet supported.")

    def get_dl_bandwidth(self, from_node: str, to_node: str) -> float:
        return self.bandwidth[from_node][to_node]

    def get_image_sizes(self, pod: Pod, arch='amd64') -> Dict[str, int]:
        """
        Returns a dictionary with the image sizes
        :param pod:
        :param arch:
        :return:
        """
        return {container.image: self.get_image_state(container.image).size[arch] for container in pod.spec.containers}



from typing import Dict, List, NamedTuple

from skippy.core.utils import parse_size_string


class ImageState:
    """
    ImageState holds runtime information about the size of container images and the number of nodes a particular image
    is deployed on. Through a manifest, a container image can be a combination of multiple images for a specific
    computing platform (amd64, arm32v7, aarch64,..) that may differ in size.
    """
    size: Dict[str, int]
    num_nodes: int = 0

    def __init__(self, size: Dict[str, int], num_nodes: int = 0):
        self.size = size
        self.num_nodes = num_nodes

    def __str__(self) -> str:
        return "ImageState%s" % self.__dict__

    def __repr__(self):
        return self.__str__()


class ResourceRequirements:
    """
    API Spec: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.13/#resourcerequirements-v1-core
    Example: https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container/

    Default values for CPU and MEM requests:
    https://github.com/kubernetes/kubernetes/blob/7f23a743e8c23ac6489340bbb34fa6f1d392db9d/pkg/scheduler/algorithm/priorities/util/non_zero.go#L31

    TODO Handling if something either limit or request is set:
    https://kubernetes.io/docs/tasks/administer-cluster/manage-resources/memory-default-namespace/
    """
    default_milli_cpu_request = 100  # 0,1 cores
    default_mem_request = 200 * 1024 * 1024  # 200 MB

    default_requests: Dict[str, float] = {"cpu": default_milli_cpu_request, "memory": default_mem_request}

    def __init__(self, requests: Dict[str, float] = None) -> None:
        super().__init__()
        self.requests = requests or dict(ResourceRequirements.default_requests)
        if self.requests.get('cpu') is None:
            self.requests['cpu'] = self.default_milli_cpu_request
        if self.requests.get('memory') is None:
            self.requests['memory'] = self.default_mem_request

    @property
    def cpu(self) -> float:
        return self.requests['cpu']

    @property
    def memory(self):
        return self.requests['memory']

    @staticmethod
    def from_str(memory: str, cpu: str):
        """
        :param memory: "64Mi"
        :param cpu: "250m"
        :return:
        """
        return ResourceRequirements({'cpu': int(cpu.rstrip('m')), 'memory': parse_size_string(memory)})


class Container:
    """
    A container is an instance of a container image that has specific resource requirements.

    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.13/#container-v1-core
    """
    resources: ResourceRequirements = ResourceRequirements()
    image: str

    def __init__(self, image: str, resources: ResourceRequirements = None) -> None:
        super().__init__()
        self.resources = resources or ResourceRequirements()
        self.image = image


class PodSpec:
    """
    A PodSpec holds the list of containers that are required by the pod, and the pod's labels.

    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.13/#podspec-v1-core
    """
    containers: List[Container]
    labels: Dict[str, str]

    def __init__(self, containers: List[Container] = None, labels: Dict[str, str] = None) -> None:
        super().__init__()
        if containers is None:
            containers = []
        if labels is None:
            labels = {}
        self.containers = containers
        self.labels = labels


class Pod:
    """
    A Pod represents a running process on your cluster.
    """
    name: str
    namespace: str
    spec: PodSpec

    def __init__(self, name: str, namespace: str, spec: PodSpec = None) -> None:
        super().__init__()
        self.name = name
        self.namespace = namespace
        self.spec = spec


class Capacity:
    """
    Node capacity
    """

    def __init__(self, cpu_millis: int = 1 * 1000, memory: int = 1024 * 1024 * 1024):
        self.memory = memory
        self.cpu_millis = cpu_millis

    def __str__(self):
        return 'Capacity(CPU: {0} Memory: {1})'.format(self.cpu_millis, self.memory)


class Node:
    """
    A node is a worker machine in Kubernetes to run pods.
    """
    name: str
    pods: List[Pod]
    capacity: Capacity
    allocatable: Capacity  # This variable is stateful and contains the *remaining* allocatable capacity
    labels: Dict[str, str]

    def __init__(self, name: str, capacity: Capacity = None, allocatable: Capacity = None,
                 labels: Dict[str, str] = None) -> None:
        super().__init__()
        self.name = name
        self.capacity = capacity or Capacity()
        self.allocatable = allocatable or Capacity()
        self.labels = labels or {}
        self.pods = list()

    def __repr__(self):
        return self.name


class SchedulingResult(NamedTuple):
    """
    What Scheduler.schedule returns.
    """
    suggested_host: Node
    feasible_nodes: int
    needed_images: List[str]  # Defines which images need to be pulled on the selected node



"""
Implementations of the kubernetes default scheduler's priority functions as well as new priority functions (Skippy).
https://github.com/kubernetes/kubernetes/tree/e318642946daab9e0330757a3556a1913bb3fc5c/pkg/scheduler/algorithm/priorities
"""
import logging
from math import fabs
from typing import Dict

from skippy.core.clustercontext import ClusterContext
from skippy.core.model import Pod, Node, Capacity, ImageState
from skippy.core.utils import normalize_image_name

logger = logging.getLogger(__name__)


def _scale_scores(scores, t_max=10):
    """
    Maps a list of measurements in range [min(scores), max(scores)] to [0, t_max].

    :param scores: the list of numbers to scale
    :param t_max: the max value to scale to
    :return: a list of mapped values
    """
    r_min = min(scores, default=0)
    r_max = max(scores, default=0)

    div = r_max - r_min

    if div == 0:
        return [0] * len(scores)

    return [int(((x - r_min) / div) * t_max) for x in scores]


def _scale_scores_inverse(scores, t_max=10):
    """
    Maps a list of measurements in range [min(scores), max(scores)] to [t_max, 0].

    :param scores: the list of numbers to scale
    :param t_max: the max value to scale to
    :return: a list of mapped values
    """
    r_min = min(scores, default=0)
    r_max = max(scores, default=0)

    div = r_min - r_max

    if div == 0:
        return [0] * len(scores)

    return [int(((x - r_max) / div) * t_max) for x in scores]


class Priority:
    """
    Abstract class for priority function implementations.
    """

    def __init__(self):
        pass

    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        """
        Calculates the score of a node for the pod.

        :param context: the cluster context
        :param pod: the pod being scheduled
        :param node: the node to score for this pod
        :return: a score value
        """
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def reduce_mapped_score(self, context: ClusterContext, pod: Pod, nodes: [Node], node_scores: [int]) -> [int]:
        """
        Reduces the already mapped scores of all nodes for a specific pod.
        The default implementation does not modify the scores anymore.
        """
        return node_scores


class EqualPriority(Priority):
    # noinspection PyMethodMayBeStatic
    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        return 1


class ImageLocalityPriority(Priority):
    """
    ImageLocalityPriority prefers nodes that have a local copy of the container images required to host the pod.
    See https://github.com/kubernetes/kubernetes/blob/v1.13.9/pkg/scheduler/algorithm/priorities/image_locality.go
    """

    mb: int = 1024 * 1024
    min_threshold: int = 23 * mb
    max_threshold: int = 1000 * mb

    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        return self.calculate_priority(context, self.sum_image_scores(context, pod, node))

    def calculate_priority(self, context: ClusterContext, sum_scores: int) -> int:
        if sum_scores < self.min_threshold:
            sum_scores = self.min_threshold
        elif sum_scores > self.max_threshold:
            sum_scores = self.max_threshold
        return int(context.max_priority * (sum_scores - self.min_threshold) / (self.max_threshold - self.min_threshold))

    def sum_image_scores(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        calc_sum = 0
        total_num_nodes = len(context.list_nodes())
        if pod.spec.containers is not Node:
            for container in pod.spec.containers:
                try:
                    image_state: ImageState = context.images_on_nodes[node.name][normalize_image_name(container.image)]
                    calc_sum += self.scaled_image_score(node, image_state, total_num_nodes)
                except KeyError:
                    pass
        return calc_sum

    # noinspection PyMethodMayBeStatic
    def scaled_image_score(self, node: Node, image_state: ImageState, total_num_nodes: int) -> int:
        spread = float(image_state.num_nodes) / float(total_num_nodes)
        return int(float(image_state.size[node.labels['beta.kubernetes.io/arch']]) * spread)


class ResourcePriority(Priority):
    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        logging.debug(f'ResourcePriority: Calculating score for {pod.name} on {node.name}')
        allocatable = node.allocatable
        requested = Capacity()
        requested.memory = 0
        requested.cpu_millis = 0
        for container in pod.spec.containers:
            requested.cpu_millis += container.resources.requests.get("cpu", container.resources.
                                                                     default_milli_cpu_request)
            requested.memory += container.resources.requests.get("memory", container.resources.default_mem_request)

        score = self.scorer(context, requested, allocatable)
        return score

    def scorer(self, context: ClusterContext, requested: Capacity, allocatable: Capacity):
        raise NotImplementedError


class BalancedResourcePriority(ResourcePriority):
    def scorer(self, context: ClusterContext, requested: Capacity, allocatable: Capacity):
        cpu_fraction = self.fraction_of_capacity(requested.cpu_millis, allocatable.cpu_millis)
        memory_fraction = self.fraction_of_capacity(requested.memory, allocatable.memory)

        # if requested >= capacity, the corresponding host should never be preferred.
        if cpu_fraction >= 1 or memory_fraction >= 1:
            return 0

        diff = fabs(cpu_fraction - memory_fraction)
        result = int((1 - diff) * float(context.max_priority))
        """
        logging.debug('BalancedResourcePriority: ')
        logging.debug(f'- requested cpu millis: {requested.cpu_millis}')
        logging.debug(f'- allocatable cpu millis: {allocatable.cpu_millis}')
        logging.debug(f'- cpu fraction: {cpu_fraction}')
        logging.debug(f'- requested memory: {requested.memory}')
        logging.debug(f'- allocatable memory: {allocatable.memory}')
        logging.debug(f'- memory fraction: {memory_fraction}')
        logging.debug(f'- diff: {diff}')
        logging.debug(f'- result: {result}')
        """
        return result

    @staticmethod
    def fraction_of_capacity(requested: int, capacity: int) -> float:
        if capacity == 0:
            capacity = 1
        return float(requested) / float(capacity)


class LocalityTypePriority(Priority):
    """
    LocalityTypePriority prefers nodes that have the locality label 'locality.skippy.io/type': 'edge'
    """

    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        # Either return the priority for the type label or 0
        priority_mapping: Dict[str, int] = {
            # Give edge nodes the highest priority
            'edge': context.max_priority,
            'cloud': 0
        }
        try:
            return priority_mapping.get(node.labels['locality.skippy.io/type'], 0)
        except KeyError:
            return 0


class CapabilityPriority(Priority):
    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        # TODO maybe we should handle capabilities like resources (where each deployment decreases the available amount)
        priority = 0
        pod_caps = dict(filter(lambda label: 'capability.skippy.io' in label[0], node.labels.items()))
        # Add 1 for each capability the pod has and the node fulfills (node affinity based on labels)
        for capability in pod_caps.items():
            if capability[0] in pod.spec.labels and capability[1] == pod.spec.labels[capability[0]]:
                priority += 1
        return priority

    def reduce_mapped_score(self, context: ClusterContext, pod: Pod, nodes: [Node], node_scores: [int]) -> [int]:
        return _scale_scores(node_scores, context.max_priority)


class LocalityPriority(Priority):
    """
    The LocalityPriority is an abstract class that can be used to implement priorities that consider the locality
    between two nodes in terms of bandwidth.
    """

    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        size = self.get_size(context, pod, node)
        target_node = self.get_target_node(context, pod, node)
        # downloading means sending from the registry means sending *from* the registry *to* the node
        bandwidth = context.get_dl_bandwidth(target_node, node.name)
        time = int(size / bandwidth)
        return time

    def reduce_mapped_score(self, context: ClusterContext, pod: Pod, nodes: [Node], node_scores: [int]) -> [int]:
        # Scale the priorities from 0 to max_priority, the lower the node score (time) the higher the priority
        # We do not adjust the values based on the minimum values.
        # Therefore f.e. if there's a download time of 10 and one of 12s, the 12s wouldn't get scored 0 but 8
        min_count_by_node_name = min(node_scores, default=0)
        max_count_by_node_name = max(node_scores, default=0)
        if max_count_by_node_name == 0:
            return [0] * len(node_scores)
        result = list(map(lambda node_count: int(context.max_priority *
                                                 (max_count_by_node_name - node_count + min_count_by_node_name) /
                                                 max_count_by_node_name),
                          node_scores))
        '''
        # Alternative:
        # Adjust to the min value, then score the download times (lowest = 10, highest = 0)
        max_count_by_node_name = max(node_scores, default=0)
        min_count_by_node_name = min(node_scores, default=0)
        if max_count_by_node_name == 0 or max_count_by_node_name == min_count_by_node_name:
            return [0] * len(node_scores)

        result = list(map(lambda node_count: int(context.max_priority *
                                                 (max_count_by_node_name - node_count) /
                                                 (max_count_by_node_name - min_count_by_node_name)), node_scores))
        '''
        return result

    def get_target_node(self, context: ClusterContext, pod: Pod, node: Node) -> str:
        raise NotImplemented()

    def get_size(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        raise NotImplemented()


class LatencyAwareImageLocalityPriority(LocalityPriority):
    """
    LatencyAwareImageLocalityPriority prefers nodes that can download the required container images more quickly. It
    does this in the same way as LocalityPriority.
    """

    def get_size(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        size = 0
        node_arch = node.labels['beta.kubernetes.io/arch']

        # determines for each container the size of the container image for the architecture of the node
        for container in pod.spec.containers:
            image_name = normalize_image_name(container.image)

            if image_name in context.images_on_nodes[node.name]:
                # node already has the image
                continue

            image_states = context.get_image_state(image_name)
            if node_arch not in image_states.size:
                replacement = list(image_states.size.keys())[0]
                logger.error("could not resolve node arch '%s' for image '%s', estimating using '%s' instead",
                             node_arch, image_name, replacement)
                node_arch = replacement

            size += context.get_image_state(image_name).size[node_arch]

        return size

    def get_target_node(self, context: ClusterContext, pod: Pod, node: Node) -> str:
        return 'registry'


class DataLocalityPriority(Priority):
    """
    DataLocalityPriority prefers nodes that are close to a storage pod that have the data items the pod is requesting.
    the pods advertise which data items they require by adding the S3 object name into a label. The priority resolves
    the closest storage node in terms of theoretically available bandwidth, but does currently not consider current
    bandwidth usage at runtime.
    """

    def map_node_score(self, context: ClusterContext, pod: Pod, node: Node) -> int:
        # FIXME: currently we assume that each function has at most one data item going in and out

        total_time = 0
        total_time += self.calculate_recv_time(context, pod, node)
        total_time += self.calculate_send_time(context, pod, node)

        return total_time

    def calculate_recv_time(self, context: ClusterContext, pod: Pod, node: Node):
        path = pod.spec.labels.get('data.skippy.io/receives-from-storage/path')

        if not path:
            return 0

        data_item = context.storage_index.stat(*path.split('/'))

        if not data_item:
            return 0

        storage_nodes = context.get_storage_nodes(path)

        # find the storage node that holds the data and has the minimal bandwidth in the required direction (down)
        min_bw_storage = None
        min_bw = float('inf')
        for storage in storage_nodes:
            if storage == node.name:
                return 0

            bandwidth = context.get_dl_bandwidth(storage, node.name)
            if bandwidth < min_bw:
                min_bw = bandwidth
                min_bw_storage = storage

        if min_bw_storage:
            return int(data_item.size / min_bw)

        return 0

    def calculate_send_time(self, context: ClusterContext, pod: Pod, node: Node):
        path = pod.spec.labels.get('data.skippy.io/sends-to-storage/path')

        if not path:
            return 0

        data_item = context.storage_index.stat(*path.split('/'))

        if not data_item:
            return 0

        storage_nodes = context.get_storage_nodes(path)

        # find the storage node that holds the data and has the minimal bandwidth in the required direction (up)
        min_bw_storage = None
        min_bw = float('inf')
        for storage in storage_nodes:
            if storage == node.name:
                return 0

            bandwidth = context.get_dl_bandwidth(node.name, storage)
            if bandwidth < min_bw:
                min_bw = bandwidth
                min_bw_storage = storage

        if min_bw_storage:
            return int(data_item.size / min_bw)

        return 0

    def reduce_mapped_score(self, context: ClusterContext, pod: Pod, nodes: [Node], node_scores: [int]) -> [int]:
        return _scale_scores_inverse(node_scores, context.max_priority)



import os

import setuptools

try:
    with open("README.md", "r") as fh:
        long_description = fh.read()
except FileNotFoundError:
    pass

try:
    with open("requirements-dev.txt", "r") as fh:
        tests_require = [line for line in fh.read().split(os.linesep) if line]
except FileNotFoundError:
    pass

with open("requirements.txt", "r") as fh:
    install_requires = [line for line in fh.read().split(os.linesep) if line]

setuptools.setup(
    name="edgerun-skippy-core",
    version="0.2.0.dev1",
    author="Alexander Rashed, Thomas Rausch, Philipp Raith",
    author_email="alexander.rashed@gmail.com, t.rausch@dsg.tuwien.ac.at, p.raith@dsg.tuwien.ac.at",
    description="Skippy scheduler core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/edgerun/skippy-core",
    download_url="https://github.com/edgerun/skippy-core",
    packages=setuptools.find_packages(),
    setup_requires=['wheel'],
    test_suite="tests",
    tests_require=tests_require,
    install_requires=install_requires,
    pyton_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
)
